from typing import Any, Optional, List, Union, Dict
from ....core.models.IModel import IModel
from ....core.toolkits.IToolkit import IToolkit
from ....core.parsers.IParser import IParser
from ....core.conversations.IConversation import IConversation
from ....core.documents.IDocument import IDocument
from ....core.retrievers.IRetriever import IRetriever
from ....core.messages import IMessage

from ..base.SwarmAgentBase import SwarmAgentBase
from ...messages.concrete import HumanMessage, AgentMessage

class ChatSwarmAgent(SwarmAgentBase):
    def __init__(self, 
                 model: IModel, 
                 conversation: Optional[IConversation] = None, 
                 toolkit: Optional[IToolkit] = None, 
                 parser: Optional[IParser] = None,
                 documents: Optional[List[IDocument]] = None, 
                 retriever: Optional[IRetriever] = None):
        super().__init__(model, conversation, toolkit, parser, documents, retriever)

    def exec(self, input_data: Union[str, IMessage], model_kwargs: Optional[Dict] = {}) -> Any:
        conversation = self.get_conversation()
        model = self.get_model()

        # Check if the input is a string, then wrap it in a HumanMessage
        if isinstance(input_data, str):
            human_message = HumanMessage(input_data)
        elif isinstance(input_data, IMessage):
            human_message = input_data
        else:
            raise TypeError("Input data must be a string or an instance of Message.")

        # Add the human message to the conversation
        conversation.add_message(human_message)
        
        # Retrieve the conversation history and predict a response
        messages = conversation.as_dict()
        if model_kwargs:
            prediction = model.predict(messages=messages, **model_kwargs)
        else:
            prediction = model.predict(messages=messages)
        # Create an AgentMessage instance with the model's response and update the conversation
        agent_message = AgentMessage(prediction)
        conversation.add_message(agent_message)
        
        return prediction


from abc import ABC, abstractmethod
from typing import List
from ....core.vector_stores.IVectorStore import IVectorStore
from ....core.document_stores.IDocumentStore import IDocumentStore
from ....core.retrievers.IRetriever import IRetriever
from ....core.documents.IDocument import IDocument

class DocumentRetrieverBase(IRetriever, ABC):
    """
    Abstract base class for Retriever implementations.
    
    This class provides the foundation for building retrievers that 
    can fetch relevant documents based on a query. It defines the basic
    structure and required methods that concrete subclasses should implement.
    """

    def __init__(self, document_store: IDocumentStore):
        """
        Initialize the RetrieverBase with a document store and a vector store.

        Args:
            document_store (IDocumentStore): An instance of a document store.
            vector_store (IVectorStore): An instance of a vector store.
        """
        self.document_store = document_store
        
    @abstractmethod
    def retrieve(self, query: str, top_k: int = 5) -> List[IDocument]:
        """
        Retrieve the top_k most relevant documents based on the given query.
        
        Args:
            query (str): The query string used for document retrieval.
            top_k (int): The number of top relevant documents to retrieve.
        
        Returns:
            List[IDocument]: A list of the top_k most relevant documents.
        """
        pass


from typing import List, Union, Optional
import numpy as np
from gensim.models.doc2vec import TaggedDocument, Doc2Vec
from swarmauri.core.document_stores.IDocumentStore import IDocumentStore
from swarmauri.core.retrievers.IRetriever import IRetriever
from swarmauri.core.documents.IDocument import IDocument
from swarmauri.standard.vector_stores.concrete.CosineDistance import CosineDistance
from swarmauri.standard.vectors.concrete.SimpleVector import SimpleVector
import gensim.downloader as api

class Doc2VecDocumentStore(IDocumentStore, IRetriever):
    def __init__(self):
        self.model = Doc2Vec(vector_size=2000, window=10, min_count=1, workers=5)  # Example parameters; adjust as needed
        self.documents = []
        self.metric = CosineDistance()

    def add_document(self, document: IDocument) -> None:
        self.documents.append(document)
        self._train()
        
    def add_documents(self, documents: List[IDocument]) -> None:
        self.documents.extend(documents)
        self._train()
        
    def get_document(self, doc_id: str) -> Union[IDocument, None]:
        for document in self.documents:
            if document.id == doc_id:
                return document
        return None
        
    def get_all_documents(self) -> List[IDocument]:
        return self.documents
        
    def delete_document(self, doc_id: str) -> None:
        self.documents = [doc for doc in self.documents if doc.id != doc_id]

    def update_document(self, doc_id: str, updated_document: IDocument) -> None:
        for i, document in enumerate(self.documents):
            if document.id == doc_id:
                self.documents[i] = updated_document
                break
        self._train()

    def retrieve(self, query: str, top_k: int = 5) -> List[IDocument]:
        """
        Retrieve documents similar to the query string based on Word2Vec embeddings.
        """
        query_vector = self.model.infer_vector(query.split())

        similarities = self.model.dv.most_similar([query_vector], topn=self.model.corpus_count)
        top_k_indices = sorted(range(len(similarities)), key=lambda i: similarities[i], reverse=True)[:top_k]
        return [self.documents[i] for i in top_k_indices]

    def _train(self):
        tagged_data = [TaggedDocument(words=_d.content.split(), tags=[str(i)]) for i, _d in enumerate(self.documents) if _d.content]
        self.model.build_vocab(tagged_data)
        self.model.train(tagged_data, total_examples=self.model.corpus_count, epochs=self.model.epochs)


from typing import List, Union, Any

import torch
from transformers import BertModel, BertTokenizer

from swarmauri.core.vectorizers.IVectorize import IVectorize
from swarmauri.core.vectors.IVector import IVector

from swarmauri.standard.vectors.concrete.SimpleVector import SimpleVector

class BERTEmbeddingVectorizer(IVectorize):
    """
    A vectorizer that generates embeddings for input text using a BERT model.
    """
    
    def __init__(self, model_name: str = 'bert-base-uncased'):
        """
        Initializes the BERTEmbeddingVectorizer with a specified BERT model.
        
        Parameters:
        - model_name (str): The name of the pre-trained BERT model to use.
        """
        self.tokenizer = BertTokenizer.from_pretrained(model_name)
        self.model = BertModel.from_pretrained(model_name)
        self.model.eval()  # Set model to evaluation mode for inference
    

    
    def fit(self, data: Union[str, Any]) -> List[IVector]:
        """
        Vectorizes input data using the BERT model to generate embeddings.

        Parameters:
        - data (Union[str, Any]): The input data to be vectorized, expected to be text.

        Returns:
        - List[IVector]: A list of IVector instances representing the generated BERT embeddings.
        """
        # Tokenization
        inputs = self.tokenizer(data, return_tensors='pt', padding=True, truncation=True, max_length=512)

        # Generate embeddings
        with torch.no_grad():
            outputs = self.model(**inputs)

        # Use the last hidden state as the embeddings (batch_size, sequence_length, hidden_size)
        embeddings = outputs.last_hidden_state

        # Convert embeddings to SimpleVector instances
        vectorized_data = []
        for embedding in embeddings:
            # Flatten the embeddings or take mean/max pooling across the sequence_length dimension if needed
            vector = SimpleVector(embedding.mean(dim=0).cpu().numpy().tolist())  # Taking mean across the token embeddings
            vectorized_data.append(vector)

        return vectorized_data
    
    def transform(self):
        raise NotImplementedError('Transform not yet implemented on BERTEmbeddingVectorizer class.')


from typing import Any, Optional, List
from abc import ABC, abstractmethod
from ....core.agents.ISwarmAgent import ISwarmAgent
from ....core.models.IModel import IModel
from ....core.toolkits.IToolkit import IToolkit
from ....core.parsers.IParser import IParser
from ....core.conversations.IConversation import IConversation
from ....core.documents.IDocument import IDocument
from ....core.retrievers.IRetriever import IRetriever
from ...messages.concrete.HumanMessage import HumanMessage

class SwarmAgentBase(ISwarmAgent, ABC):
    @abstractmethod
    def __init__(self, 
                 model: IModel, 
                 conversation: Optional[IConversation] = None,
                 toolkit: Optional[IToolkit] = None, parser: Optional[IParser] = None, 
                 documents: Optional[List[IDocument]] = [], retriever: Optional[IRetriever] = None):

        super().__init__(model, conversation, toolkit, parser, documents, retriever)
        self._model = model
        self._toolkit = toolkit
        self._parser = parser
        self._conversation = conversation
        self._documents = documents 
        self._retriever = retriever

    def exec(self, input_str: Optional[Any]) -> Any:
        pass
    
    @property
    def model(self) -> IModel:
        return self._model
    
    @model.setter
    def model(self, value) -> IModel:
        self._model = value        

    def get_model(self) -> IModel:
        return self._model
    
    def get_toolkit(self) -> Optional[IToolkit]:
        return self._toolkit
    
    @property
    def parser(self) -> IModel:
        return self._parser
    
    @model.setter
    def parser(self, value) -> IModel:
        self._parser = value
    
    
    def get_parser(self) -> Optional[IParser]:
        return self._parser

    @property
    def conversation(self) -> Optional[IConversation]:
        return self._conversation

    @conversation.setter
    def conversation(self, value) -> Optional[IConversation]:
        self._conversation = value

    def get_conversation(self) -> Optional[IConversation]:
        return self._conversation
    
    def get_documents(self) -> List[IDocument]:
        return self._documents
    
    def get_retriever(self) -> Optional[IRetriever]:
        return self._retriever 
    
    def __getattr__(self, name):
        # Example of transforming attribute name from simplified to internal naming convention
        internal_name = f"_{name}"
        if internal_name in self.__dict__:
            return self.__dict__[internal_name]
        raise AttributeError(f"'{self.__class__.__name__}' object has no attribute '{name}'")
    
    def __setattr__(self, name, value):
        # Direct assignment to the __dict__ to bypass any potential infinite recursion
        # from setting attributes that do not explicitly exist.
        object.__setattr__(self, name, value) 
        
        
    def __str__(self):
        class_name = self.__class__.__name__
        variables_str = ", ".join(f"{k}={v}" for k, v in self.__dict__.items())
        return f"<{class_name} {variables_str}>"
        
    def __repr__(self):
        class_name = self.__class__.__name__
        variables_str = ", ".join(f"{k}={v}" for k, v in self.__dict__.items())
        return f"{class_name} ({variables_str})"


from typing import Any, Optional, List, Union, Dict
import json

from ....core.models.IModel import IModel
from ....core.toolkits.IToolkit import IToolkit
from ....core.parsers.IParser import IParser
from ....core.conversations.IConversation import IConversation
from ....core.documents.IDocument import IDocument
from ....core.retrievers.IRetriever import IRetriever
from ....core.messages import IMessage

from ..base.SwarmAgentBase import SwarmAgentBase
from ...messages.concrete import HumanMessage, AgentMessage, FunctionMessage


class MultiPartyToolAgent(SwarmAgentBase):
    def __init__(self, 
                 model: IModel, 
                 conversation: Optional[IConversation] = None, 
                 toolkit: Optional[IToolkit] = None, 
                 parser: Optional[IParser] = None,
                 documents: Optional[List[IDocument]] = None, 
                 retriever: Optional[IDocument] = None,
                 name: str = None):
        super().__init__(model, conversation, toolkit, parser, documents, retriever)
        self.name = name

    def exec(self, input_data: Union[str, IMessage], model_kwargs: Optional[Dict] = {}) -> Any:
        conversation = self.get_conversation()
        model = self.get_model()
        tools = self.get_toolkit()
        

        # Check if the input is a string, then wrap it in a HumanMessage
        if isinstance(input_data, str):
            human_message = HumanMessage(input_data)
        elif isinstance(input_data, IMessage):
            human_message = input_data
        else:
            raise TypeError("Input data must be a string or an instance of Message.")

        if input_data != "":
            # Add the human message to the conversation
            conversation.add_message(human_message, sender_id=self.name)
            
        
        # Retrieve the conversation history and predict a response
        messages = conversation.as_dict()
        

        if model_kwargs:
            prediction = model.predict(messages=messages, 
                                   tools=tools.tools, 
                                   tool_choice="auto",
                                   **model_kwargs)
        else:
            prediction = model.predict(messages=messages)
        
        
        prediction_message = prediction.choices[0].message
        agent_response = prediction_message.content
        
        agent_message = AgentMessage(content=prediction_message.content, 
                                     tool_calls=prediction_message.tool_calls)
        conversation.add_message(agent_message, sender_id=self.name)
        
        tool_calls = prediction.choices[0].message.tool_calls
        if tool_calls:
        
            for tool_call in tool_calls:
                func_name = tool_call.function.name
                
                func_call = self.toolkit.get_tool_by_name(func_name)
                func_args = json.loads(tool_call.function.arguments)
                func_result = func_call(**func_args)
                
                func_message = FunctionMessage(func_result, 
                                               name=func_name, 
                                               tool_call_id=tool_call.id)
                conversation.add_message(func_message, sender_id=self.name)
            
            
            messages = conversation.as_dict()
            rag_prediction = model.predict(messages=messages, 
                                           tools=tools.tools, 
                                           tool_choice="none")
            
            prediction_message = rag_prediction.choices[0].message
            
            agent_response = prediction_message.content
            if agent_response != "":
                agent_message = AgentMessage(agent_response)
                conversation.add_message(agent_message, sender_id=self.name)
            prediction = rag_prediction
            
        return agent_response 
    


from sklearn.feature_extraction.text import TfidfVectorizer as SklearnTfidfVectorizer
from typing import List, Union, Any
from swarmauri.core.vectorizers.IVectorize import IVectorize
from swarmauri.core.vectors.IVector import IVector
from swarmauri.standard.vectors.concrete.SimpleVector import SimpleVector

class TFIDFVectorizer(IVectorize):
    def __init__(self):
        # Using scikit-learn's TfidfVectorizer as the underlying mechanism
        self.tfidf_vectorizer = SklearnTfidfVectorizer()
        super().__init__()
        
    def extract_features(self):
        return self.tfidf_vectorizer.get_feature_names_out()

    def fit(self, data: Union[str, Any]) -> List[IVector]:
        """
        Vectorizes the input data using the TF-IDF scheme.

        Parameters:
        - data (Union[str, Any]): The input data to be vectorized. Expected to be a single string (document)
                                  or a list of strings (corpus).

        Returns:
        - List[IVector]: A list containing IVector instances, each representing a document's TF-IDF vector.
        """
        if isinstance(data, str):
            data = [data]  # Convert a single string into a list for the vectorizer
        
        tfidf_matrix = self.tfidf_vectorizer.fit_transform(data)

        # Convert the sparse matrix rows into SimpleVector instances
        vectors = [SimpleVector(vector.toarray().flatten()) for vector in tfidf_matrix]

        return vectors
    
    def transform(self, data: Union[str, Any]) -> List[IVector]:
        """
        Vectorizes the input data using the TF-IDF scheme.

        Parameters:
        - data (Union[str, Any]): The input data to be vectorized. Expected to be a single string (document)
                                  or a list of strings (corpus).

        Returns:
        - List[IVector]: A list containing IVector instances, each representing a document's TF-IDF vector.
        """
        if isinstance(data, str):
            data = [data]  # Convert a single string into a list for the vectorizer
        
        tfidf_matrix = self.tfidf_vectorizer.fit_transform(data)

        # Convert the sparse matrix rows into SimpleVector instances
        vectors = [SimpleVector(vector.toarray().flatten()) for vector in tfidf_matrix]

        return vectors


from typing import Any, Optional, List

from ....core.models.IModel import IModel
from ....core.toolkits.IToolkit import IToolkit
from ....core.parsers.IParser import IParser
from ....core.conversations.IConversation import IConversation
from ....core.documents.IDocument import IDocument
from ....core.retrievers.IRetriever import IRetriever
from ....core.messages import IMessage

from ..base.SwarmAgentBase import SwarmAgentBase
from ...messages.concrete import HumanMessage

class SingleCommandAgent(SwarmAgentBase):
    def __init__(self, model: IModel, 
                 conversation: Optional[IConversation] = None, 
                 toolkit: Optional[IToolkit] = None, 
                 parser: Optional[IParser] = None,
                 documents: Optional[List[IDocument]] = None, 
                 retriever: Optional[IRetriever] = None):
        super().__init__(model, conversation, toolkit, parser, documents, retriever)

    def exec(self, input_str: Optional[str] = None) -> Any:
        model = self.get_model()
        prediction = model.predict(input_str)
        
        return prediction


from abc import ABC, abstractmethod
from typing import Any, Optional, List
from ..models.IModel import IModel
from ..toolkits.IToolkit import IToolkit
from ..parsers.IParser import IParser
from ..conversations.IConversation import IConversation
from ..documents.IDocument import IDocument
from ..retrievers.IRetriever import IRetriever
from ..messages.IMessage import IMessage

class ISwarmAgent(ABC):
    
    @abstractmethod
    def __init__(self, model: IModel, 
                 conversation: Optional[IConversation] = None, 
                 toolkit: Optional[IToolkit] = None, 
                 parser: Optional[IParser] = None, 
                 documents: Optional[List[IDocument]] = [], 
                 retriever: Optional[IRetriever] = None):
        pass
    
    @abstractmethod
    def exec(self, input_str: Optional[Any]) -> Any:
        pass
    
    @abstractmethod
    def get_model(self) -> IModel:
        """Returns the model component of the agent."""
        pass
    
    @abstractmethod
    def get_toolkit(self) -> Optional[IToolkit]:
        """Returns the toolkit component of the agent, if available."""
        pass
    
    @abstractmethod
    def get_parser(self) -> Optional[IParser]:
        """Returns the parser component of the agent, if available."""
        pass

    @abstractmethod
    def get_conversation(self) -> Optional[IConversation]:
        """Returns the conversation manager component of the agent, if available."""
        pass
    
    @abstractmethod
    def get_documents(self) -> List[IDocument]:
        """Returns the list of document components associated with the agent."""
        pass
    
    @abstractmethod
    def get_retriever(self) -> Optional[IRetriever]:
        """Returns the retriever component of the agent, if available."""
        pass

